#pragma once
#include "myincludes.hh"
//#include <dune/xt/grid/grids.hh>
//#include <dune/xt/grid/view/periodic.hh>
#include <dune/pdelab/localoperator/stokesparameter.hh>
#include <dune/pdelab/gridfunctionspace/vectorgridfunctionspace.hh>
#include <dune/pdelab/localoperator/taylorhoodnavierstokes.hh>
#include <dune/pdelab/backend/istl/seqistlsolverbackend.hh>
#include <dune/pdelab/adaptivity/adaptivity.hh>

//#include "acfs_fem_allencahn.hh"
//#include "acfs_fem_cahnhilliard.hh"
//#include "acfs_fem_navierstokes.hh"
//#include "acfs_fem_ac_estimator.hh"
//#include "threephase_fem_cahnhilliard.hh"
//#include "threephase_fem_navierstokes.hh"
//#include "threephase_dgf_coloring.hh"
#include <dune/phasefield/localoperator/chns_fem_cahnhilliard.hh>
#include <dune/phasefield/localoperator/chns_fem_navierstokes.hh>

#include <dune/phasefield/acfs_problem.hh>
#include <dune/phasefield/acfs_ibvp.hh>
#include <dune/phasefield/utility/ISTLBackend_NOVLP_BASE_PREC_FOR_GMRES.hh>


//template <class A>
//using MySolver = Dune::RestartedGMResSolver<A, A, A>;


//_________________________________________________________________________________________________________________________

template<typename Grid>
class chns_base
{
public:
  typedef typename Grid::LeafGridView GV;
  const int dim = GV::dimension;
  typedef double RF;


  chns_base(Grid& grid, Dune::ParameterTree& ptree)
  {
  }

};

class TimestepManager
{
public:
  double T;
  double dt;
  double t;

  double dtMin;
  double dtMax;
  double dtIncreaseFactor;
  double dtDecreaseFactor;

  double lastsave;
  double saveintervall;

  int verb;

  TimestepManager(const Dune::ParameterTree & param, int verbosity)
  {
    verb = verbosity;
    T=param.get("tMax",(double)1.0);
    dt=param.get("dt",(double)0.1);

    dtMin = param.get("dtMin",(double)dt);
    dtMax = param.get("dtMax",(double)dt);
    dtIncreaseFactor = param.get("dtIncreaseFactor",(double)1.2);
    dtDecreaseFactor = param.get("dtDecreaseFactor",(double)0.5);

    saveintervall = param.get("temporal.saveintervall",(double)0.0);

    t=param.get("tIni",(double)0.0);
    lastsave = t - saveintervall;
  }

  // RETURN VALUE: -1 -> ABORT, 0-> REDO TIMESTEP, 1-> OK
  template<typename NS_NEWTON, typename CH_NEWTON, typename VTKWRITER>
  int doTimestep(NS_NEWTON& ns_newton, CH_NEWTON& ch_newton, VTKWRITER& vtkwriter)
  {
    try
    {
      if (verb > 1)   std::cout << "= Begin newtonapply CH:" << std::endl;
      ch_newton.apply();

      if (verb > 1)   std::cout << "= Begin newtonapply NS:" << std::endl;
      ns_newton.apply();
    }
    catch(...)
    {
      if(dt <= dtMin)
      {
        std::cout << "===============================" << std::endl;
        if(dtMin==dtMax)
          std::cout << "Problem with Newton Solver, aborting :(" << std::endl;
        else
          std::cout << "Timestep too small, aborting :(" << std::endl;
        std::cout << "===============================" << std::endl;
        return -1;
      }
      dt = std::max(dt*dtDecreaseFactor,dtMin);
      if (verb > 0) std::cout << "Reducing Timestep to "<< dt << std::endl;

      return 0;
    }

    t += dt;
    if(dt < dtMax)
    {
      dt = std::min(dt*dtIncreaseFactor, dtMax);
      if (verb > 0)   std::cout << "New Timestep "<< dt << std::endl;
    }

    if( t>lastsave+saveintervall )
    {
        lastsave = t;
        vtkwriter.write(t,Dune::VTK::ascii);
    }
    return 1;
  }


  bool isFinished()
  {
    return (t>=T);
  }
};

template<typename GFS, typename NS_BDRY_FN, typename CH_BDRY_FN, typename NS_V, typename CH_V>
void interpolateToBdry(const GFS& gfs, const NS_BDRY_FN& nsBdryFn, const CH_BDRY_FN& chBdryFn, NS_V& nsV, CH_V& chV)
{
  NS_V nsInterp(gfs.ns);
  CH_V chInterp(gfs.ch);
  Dune::PDELab::interpolate(nsBdryFn, gfs.ns, nsInterp);
  Dune::PDELab::interpolate(chBdryFn, gfs.ch, chInterp);
  Dune::PDELab::copy_constrained_dofs(gfs.nsCC, nsInterp, nsV);
  Dune::PDELab::copy_constrained_dofs(gfs.chCC, chInterp, chV);
}

//_________________________________________________________________________________________________________________________

template<typename RF, typename ES, typename V_FEM, typename P_FEM, typename PHI_FEM>
class GFS_2f0s
{
public:
  typedef Dune::PDELab::ConformingDirichletConstraints ConstraintsAssembler;
  typedef Dune::PDELab::ISTL::VectorBackend<> VectorBackend;
  typedef Dune::PDELab::ISTL::VectorBackend<Dune::PDELab::ISTL::Blocking::fixed> VB_Block;
  typedef Dune::PDELab::LexicographicOrderingTag OrderingTagL;
  typedef Dune::PDELab::EntityBlockedOrderingTag OrderingTagB;

  ConstraintsAssembler conp, conphi, conmu;

  typedef Dune::PDELab::VectorGridFunctionSpace<ES,V_FEM,ES::GridView::dimension,
                                                VectorBackend,VectorBackend,ConstraintsAssembler> V_GFS;
  V_GFS vGfs;

  typedef Dune::PDELab::GridFunctionSpace<ES, P_FEM, ConstraintsAssembler, VectorBackend> P_GFS;
  P_GFS pGfs;

  typedef Dune::PDELab::GridFunctionSpace<ES, PHI_FEM, ConstraintsAssembler, VectorBackend> PHI_GFS;
  PHI_GFS phiGfs;

  typedef Dune::PDELab::GridFunctionSpace<ES, PHI_FEM, ConstraintsAssembler, VectorBackend> MU_GFS;
  MU_GFS muGfs;

  typedef Dune::PDELab::CompositeGridFunctionSpace<VectorBackend, OrderingTagL, V_GFS, P_GFS> NS;
  NS ns;

  typedef Dune::PDELab::CompositeGridFunctionSpace<VB_Block, OrderingTagB, PHI_GFS, MU_GFS> CH;
  CH ch;

  typedef typename NS::template ConstraintsContainer<RF>::Type NS_CC;
  NS_CC nsCC;

  typedef typename CH::template ConstraintsContainer<RF>::Type CH_CC;
  CH_CC chCC;

  GFS_2f0s(const ES& es, const V_FEM& vFem, const P_FEM& pFem, const PHI_FEM& phiFem) :
      conp(),conphi(),conmu(),
      vGfs(es,vFem),
      pGfs(es,pFem, conp),
      phiGfs(es,phiFem, conphi),
      muGfs(es,phiFem, conmu),
      ns(vGfs, pGfs),
      ch(phiGfs, muGfs),
      nsCC(),
      chCC()
  {
    vGfs.name("velocity");
    pGfs.name("pressure");
    phiGfs.name("phasefield");
    muGfs.name("potential");
  }

  template<typename Bdry>
  void updateConstraintsContainer(const Bdry& bdry)
  {
    Dune::PDELab::constraints(bdry.bdryNS,ns,nsCC, 0);
    Dune::PDELab::constraints(bdry.bdryCH,ch,chCC, 0);
  }
};

//_________________________________________________________________________________________________________________________


template<typename GFS, typename NS_V, typename CH_V>
class DGF_2f0s
{
public:

  typedef Dune::PDELab::GridFunctionSubSpace<typename GFS::NS,Dune::TypeTree::TreePath<0> > VSubGFS;
  VSubGFS vSubGfs;
  typedef Dune::PDELab::VectorDiscreteGridFunction<VSubGFS,NS_V> VDGF;
  VDGF vDgf;

  typedef Dune::PDELab::GridFunctionSubSpace<typename GFS::NS,Dune::TypeTree::TreePath<1> > PSubGFS;
  PSubGFS pSubGfs;
  typedef Dune::PDELab::DiscreteGridFunction<PSubGFS,NS_V> PDGF;
  PDGF pDgf;

  typedef Dune::PDELab::GridFunctionSubSpace<typename GFS::CH,Dune::TypeTree::TreePath<0> > PhiSubGFS;
  PhiSubGFS phiSubGfs;
  typedef Dune::PDELab::DiscreteGridFunction<PhiSubGFS,CH_V> PhiDGF;
  PhiDGF phiDgf;

  typedef Dune::PDELab::GridFunctionSubSpace<typename GFS::CH,Dune::TypeTree::TreePath<1> > MuSubGFS;
  MuSubGFS muSubGfs;
  typedef Dune::PDELab::DiscreteGridFunction<MuSubGFS,CH_V> MuDGF;
  MuDGF muDgf;


  DGF_2f0s(const GFS& gfs, const NS_V& nsV, const CH_V& chV) :
      vSubGfs(gfs.ns),
      vDgf(vSubGfs,nsV),
      pSubGfs(gfs.ns),
      pDgf(pSubGfs,nsV),
      phiSubGfs(gfs.ch),
      phiDgf(phiSubGfs,chV),
      muSubGfs(gfs.ch),
      muDgf(muSubGfs,chV)
  {
  }

  template<typename VTKWRITER>
  void addToVTKWriter(VTKWRITER& vtkwriter)
  {
    vtkwriter.addVertexData(std::make_shared<Dune::PDELab::VTKGridFunctionAdapter<VDGF> >(vDgf,"v"));
    vtkwriter.addVertexData(std::make_shared<Dune::PDELab::VTKGridFunctionAdapter<PDGF> >(pDgf,"p"));
    vtkwriter.addVertexData(std::make_shared<Dune::PDELab::VTKGridFunctionAdapter<PhiDGF> >(phiDgf,"phi"));
    vtkwriter.addVertexData(std::make_shared<Dune::PDELab::VTKGridFunctionAdapter<MuDGF> >(muDgf,"mu"));
  }
};

//_________________________________________________________________________________________________________________________

template<typename IniV, typename IniP, typename IniPhi, typename IniMu>
class Ini_2f0s
{
public:
  IniV iniV;
  IniP iniP;
  IniPhi iniPhi;
  IniMu iniMu;
  typedef Dune::PDELab::CompositeGridFunction<IniV,IniP> IniNS;
  typedef Dune::PDELab::CompositeGridFunction<IniPhi,IniMu> IniCH;
  IniNS iniNS;
  IniCH iniCH;

  template<typename GV, typename Params>
  Ini_2f0s(const GV& gv, const Params& params) :
      iniV(gv, params),
      iniP(gv, params),
      iniPhi(gv, params),
      iniMu(gv, params),
      iniNS(iniV,iniP),
      iniCH(iniPhi,iniMu)
  {
  }
};

//_________________________________________________________________________________________________________________________

template<const int dim, typename BdryV_Scalar, typename BdryP, typename BdryPhi, typename BdryMu>
class Bdry_2f0s
{
public:
  BdryV_Scalar bdryV_Scalar;
  BdryP bdryP;
  BdryPhi bdryPhi;
  BdryMu bdryMu;
  typedef Dune::PDELab::PowerConstraintsParameters<BdryV_Scalar,dim>    BdryV;
  typedef Dune::PDELab::CompositeConstraintsParameters<BdryV, BdryP> BdryNS;
  typedef Dune::PDELab::CompositeConstraintsParameters<BdryPhi, BdryMu> BdryCH;
  BdryV bdryV;
  BdryNS bdryNS;
  BdryCH bdryCH;

  Bdry_2f0s() :
      bdryV_Scalar(),
      bdryP(),
      bdryPhi(),
      bdryMu(),
      bdryV(bdryV_Scalar),
      bdryNS(bdryV,bdryP),
      bdryCH(bdryPhi,bdryMu)
  {
  }
};

//_________________________________________________________________________________________________________________________

template<typename GV, typename RF, typename EST_LOP, typename CH_GFS>
class RefinementIndicator
{
public:
  //Estimator GFS
  typedef Dune::PDELab::P0LocalFiniteElementMap<RF,RF,GV::dimension> P0FEM;
  P0FEM p0fem;
  typedef Dune::PDELab::NoConstraints NCON;
  typedef Dune::PDELab::GridFunctionSpace<GV,P0FEM,NCON,Dune::PDELab::ISTL::VectorBackend<> > P0GFS;
  P0GFS p0Gfs;
  //Estimator vector
  using Z0 = Dune::PDELab::Backend::Vector<P0GFS,RF>;
  Z0 z0;

  //typedef Dune::PDELab::DiscreteGridFunction<P0GFS,Z0> Z0DGF;
  //Z0DGF z0dgf;

  //Estimator Operator
  EST_LOP estlop;
  typedef Dune::PDELab::EmptyTransformation NCC;
  typedef Dune::PDELab::ISTL::BCRSMatrixBackend<> MBE;
  typedef Dune::PDELab::GridOperator<CH_GFS,P0GFS,EST_LOP,MBE,RF,RF,RF,NCC,NCC> ESTGO;
  ESTGO estgo;

  double eta_refine;
  double eta_coarsen;
  int Rmaxlevel;
  int Rminlevel;

  RefinementIndicator(const GV& gv, const CH_GFS chGfs, const Dune::ParameterTree & param) :
    p0fem(Dune::GeometryTypes::simplex(GV::dimension)),
    p0Gfs(gv,p0fem),
    z0(p0Gfs,0.0),
    //z0dgf(p0Gfs,z0),
    estlop(),
    estgo(chGfs,p0Gfs,estlop,MBE(1))
  {
    eta_refine = param.get("etaRefine",(double)0.15);
    eta_coarsen = param.get("etaCoarsen",(double)0.05);
    Rmaxlevel = param.get("maxRefineLevel",(double)2);
    Rminlevel = param.get("minRefineLevel",(double)0);
  }

  template<typename CH_V>
  const Z0& calculateIndicator(const CH_V& chV)
  {
    estgo.residual(chV,z0);
    return z0;
  }

  template<typename Grid>
  void markGrid(Grid &grid, int verbosity=0)
  {
    if(verbosity>0) std::cout<<"mark elements"<<std::endl;
    Dune::PDELab::mark_grid(grid,z0,eta_refine,eta_coarsen, Rminlevel,Rmaxlevel,verbosity);
  }
};
