
#pragma once
#include "myincludes.hh"
#include <cmath>


template<typename GV, typename RF, std::size_t dim_range, typename Params>
class ZeroInitialVector :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim_range>,ZeroInitialVector<GV,RF,dim_range, Params> >
{
public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim_range> Traits;

  ZeroInitialVector(const GV & gv, const Params & params) :
    Dune::PDELab::AnalyticGridFunctionBase<Traits, ZeroInitialVector>(gv) {}

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    y=0;
  }
};

//______________________________________________________________________________________________________________________

template<typename GV, typename RF, typename Params>
class ZeroInitial
  : public ZeroInitialVector<GV,RF,1,Params>
{
public:
  ZeroInitial(const GV & gv, const Params & params) : ZeroInitialVector<GV,RF,1,Params>(gv,params) {}
};

//_________________________________________________________________________________________________________________________

template<typename GV, typename RF, typename Params>
class PhiInitial_RayleighTaylor :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_RayleighTaylor<GV,RF,Params> >
{
  RF eps;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_RayleighTaylor(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_RayleighTaylor<GV,RF,Params> > (gv)
     {
       eps = params.eps;
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    y = 0.5*(1-tanh(1/eps/sqrt(2)*(x[1]-0.75-0.01*cos(x[0]*2*3.1415))));
  }
};

//_________________________________________________________________________________________________________________________
