#pragma once

#include "myincludes.hh"

#include<dune/pdelab/common/quadraturerules.hh>
#include<dune/pdelab/finiteelement/localbasiscache.hh>


template<typename CHGFS>
class ACFS_FEM_AC_Estimator :
  public Dune::PDELab::LocalOperatorDefaultFlags
{
private:
  typedef typename CHGFS::template Child<0>::Type FEM;
  typedef typename FEM::Traits::FiniteElementType::Traits::LocalBasisType LocalBasis;
  Dune::PDELab::LocalBasisCache<LocalBasis> cache;
  enum {dim=LocalBasis::Traits::dimDomain};

public:
  enum {doPatternVolume=false};
  //enum {doLambdaVolume=true};
  enum {doAlphaVolume=true};

  //constructor
  ACFS_FEM_AC_Estimator()
  {
  }


  template<typename EG, typename LFSU, typename X, typename LFSV, typename R>
  void alpha_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv, R& r) const
  {
    using namespace Dune::Indices;
    const auto& pfspace = child(lfsu,_0);
    typedef decltype(makeZeroBasisFieldValue(pfspace)) RF;

    auto geo = eg.geometry();
    const auto& rElement = Dune::ReferenceElements<double, dim>::general(geo.type());
    RF umax=0.0;
    RF umin=0.0;
  //  Dune::FieldVector<RF, dim> xg = eg.geometry().global( rElement.position(0, dim) );
    for (int i=0; i<geo.corners(); i++)
    {
      RF u=0.0;
      auto& phihat = cache.evaluateFunction(rElement.position(i, dim),pfspace.finiteElement().localBasis());
      for (size_t j=0; j<pfspace.size(); j++)
        u += x(pfspace,j)*phihat[j];

      if(i==0)
      {
        umax = u;
        umin = u;
      }
      else
      {
        if(u>umax) umax = u;
        if(u<umin) umin = u;
      }
      //xg = eg.geometry().global( rElement.position(i, dim) );
    }
    r.accumulate(lfsv,0,(umax-umin));//(xg[0]>0.25&&xg[0]<0.75)? (umax-umin):0 );
  }

};
