#pragma once
#include "myincludes.hh"
//#include <dune/xt/grid/grids.hh>
//#include <dune/xt/grid/view/periodic.hh>
#include <dune/pdelab/localoperator/stokesparameter.hh>
#include <dune/pdelab/gridfunctionspace/vectorgridfunctionspace.hh>
#include <dune/pdelab/localoperator/taylorhoodnavierstokes.hh>
#include <dune/pdelab/backend/istl/seqistlsolverbackend.hh>
#include <dune/pdelab/adaptivity/adaptivity.hh>

//#include "acfs_fem_allencahn.hh"
//#include "acfs_fem_cahnhilliard.hh"
//#include "acfs_fem_navierstokes.hh"
#include "acfs_fem_ac_estimator.hh"
//#include "threephase_fem_cahnhilliard.hh"
//#include "threephase_fem_navierstokes.hh"
//#include "threephase_dgf_coloring.hh"
#include <dune/phasefield/localoperator/chns_fem_cahnhilliard.hh>
#include <dune/phasefield/localoperator/chns_fem_navierstokes.hh>

#include <dune/phasefield/acfs_problem.hh>
#include <dune/phasefield/acfs_ibvp.hh>
#include <dune/phasefield/utility/ISTLBackend_NOVLP_BASE_PREC_FOR_GMRES.hh>

#include <dune/phasefield/chns_base.hh>
#include <dune/phasefield/initial_fn.hh>
#include <dune/phasefield/debug/vector_debug.hh>
//template <class A>
//using MySolver = Dune::RestartedGMResSolver<A, A, A>;


//_________________________________________________________________________________________________________________________

template<typename Grid, typename ES, typename V_FEM, typename P_FEM, typename PHI_FEM>
void chns_driver (Grid& grid, const ES& es, const V_FEM& vFem, const P_FEM pFem, const PHI_FEM& phiFem, Dune::ParameterTree& ptree, Dune::MPIHelper& helper)
{

  typedef typename Grid::LeafGridView GV;
  //std::bitset<2> periodic_directions(std::string("01"));
  //typedef Dune::XT::Grid::PeriodicGridView<LGV> GV;
  GV gv = grid.leafGridView();
  // dimension and important types
  const int dim = GV::dimension;
  typedef double RF;                   // type for computations



  RF initialvmax = ptree.get("initial.vmax",(RF)1);
  RF vmaxtime = ptree.get("initial.vmaxtime",(RF)1);


  TimestepManager timestepManager(ptree.sub("Time"),2);

  Problem<RF> problem(ptree.get("temporal.tau",(RF)0.1),ptree);
  problem.setTime(timestepManager.t);


//_________________________________________________________________________________________________________________________
// Make grid function spaces


  typedef GFS_2f0s<RF, ES, V_FEM, P_FEM, PHI_FEM> GFS;
  GFS gfs(es,vFem,pFem,phiFem);
  typedef typename GFS::NS NS_GFS;
  typedef typename GFS::CH CH_GFS;


  //_________________________________________________________________________________________________________________________
  //Constraints

  typedef Bdry_2f0s<dim, Dune::PDELab::DirichletConstraintsParameters,
              Dune::PDELab::NoDirichletConstraintsParameters,
              Dune::PDELab::NoDirichletConstraintsParameters,
              Dune::PDELab::NoDirichletConstraintsParameters> Bdry;
  Bdry bdry;
  gfs.updateConstraintsContainer(bdry);

  //_________________________________________________________________________________________________________________________
  //Coefficient vectors

  using CH_V = Dune::PDELab::Backend::Vector<CH_GFS, RF>;
  using NS_V = Dune::PDELab::Backend::Vector<NS_GFS, RF>;
  CH_V chV(gfs.ch);
  NS_V nsV(gfs.ns);
  CH_V chVOld = chV;
  NS_V nsVOld = nsV;

  //_________________________________________________________________________________________________________________________
  //Initial Values

    typedef Ini_2f0s< ZeroInitialVector<GV,RF,dim,Problem<RF>>,
                      ZeroInitial<GV,RF,Problem<RF>>,
                      PhiInitial_RayleighTaylor<GV,RF,Problem<RF>>,
                      ZeroInitial<GV,RF,Problem<RF>> >  Initial;
    Initial initial(gv,problem);
    Dune::PDELab::interpolate(initial.iniCH , gfs.ch, chV);
    Dune::PDELab::interpolate(initial.iniNS , gfs.ns, nsV);

  //_________________________________________________________________________________________________________________________
  // LocalOperators

  typedef CHNS_FEM_NavierStokes<Problem<RF>, CH_GFS, CH_V, NS_GFS, NS_V> NS_LOP;
  NS_LOP nsLop(problem, gfs.ch, chV, chVOld, gfs.ns, nsVOld);

  typedef CHNS_FEM_CahnHilliard<Problem<RF>, CH_GFS, CH_V, NS_GFS, NS_V> CH_LOP;
  CH_LOP chLop(problem, gfs.ch, chVOld, gfs.ns, nsV);

  typedef Dune::PDELab::ISTL::BCRSMatrixBackend<> MBE;
  MBE mbe(10);// Maximal number of nonzeroes per row can be cross-checked by patternStatistics().


  //_________________________________________________________________________________________________________________________
  // Construct discrete Grid Functions
  typedef DGF_2f0s<GFS, NS_V, CH_V> DGF;
  DGF dgf(gfs, nsV, chV);

  //_________________________________________________________________________________________________________________________
  // prepare VTK writer
  std::string filename=ptree.get("output.filename","output");
  struct stat st;
  if( stat( filename.c_str(), &st ) != 0 )
    {
      int stat = 0;
      stat = mkdir(filename.c_str(),S_IRWXU|S_IRWXG|S_IRWXO);
      if( stat != 0 && stat != -1)
        std::cout << "Error: Cannot create directory "
                  << filename << std::endl;
    }
  auto stationaryvtkwriter = std::make_shared<Dune::SubsamplingVTKWriter<GV> >(gv,0);
  Dune::VTKSequenceWriter<GV> vtkwriter(stationaryvtkwriter,filename.c_str(),filename.c_str(),"");
  dgf.addToVTKWriter(vtkwriter);

  vtkwriter.write(timestepManager.t,Dune::VTK::ascii);

//_________________________________________________________________________________________________________________________
  printVector(chV,50,"chV");
  printVector(nsV,50,"nsV");


  //____________________________________________________________________________________________________________________
  //int i_it=0;
  chVOld = chV;
  nsVOld = nsV;
  while (!timestepManager.isFinished())
    {
      problem.setTime(timestepManager.t);

      // Grid Operators
      typedef Dune::PDELab::GridOperator<NS_GFS,NS_GFS,NS_LOP,MBE,RF,RF,RF,typename GFS::NS_CC,typename GFS::NS_CC> NS_GO;
      NS_GO nsGo(gfs.ns,gfs.nsCC,gfs.ns,gfs.nsCC,nsLop,mbe);

      typedef Dune::PDELab::GridOperator<CH_GFS,CH_GFS,CH_LOP,MBE,RF,RF,RF,typename GFS::CH_CC,typename GFS::CH_CC> CH_GO;
      CH_GO chGo(gfs.ch,gfs.chCC,gfs.ch,gfs.chCC,chLop,mbe);

      // Linear solver
      typedef ISTLBackend_NOVLP_BASE_PREC_FOR_GMRES
                  <CH_GO, Dune::SeqILU0, Dune::RestartedGMResSolver> CH_LS; CH_LS ls_ch(chGo,500,3,1);

      typedef ISTLBackend_NOVLP_BASE_PREC_FOR_GMRES
                  <NS_GO, Dune::SeqILU0, Dune::RestartedGMResSolver> NS_LS; NS_LS ls_ns(nsGo,500,3,1);


      typedef Dune::PDELab::Newton<NS_GO,NS_LS,NS_V> PDESolver1;
      PDESolver1 ns_newton(nsGo,nsV,ls_ns);
      ns_newton.setReassembleThreshold(0.0);
      ns_newton.setVerbosityLevel(2);
      ns_newton.setMaxIterations(50);
      ns_newton.setLineSearchMaxIterations(30);

      ns_newton.setAbsoluteLimit(1e-9);
      ns_newton.setReduction(1e-7);

      typedef Dune::PDELab::Newton<CH_GO,CH_LS,CH_V> PDESolver2;
      PDESolver2 ch_newton(chGo,chV,ls_ch);
      ch_newton.setReassembleThreshold(0.0);
      ch_newton.setVerbosityLevel(1);
      ch_newton.setMaxIterations(50);
      ch_newton.setLineSearchMaxIterations(50);



      //_________________________________________________________________________________________________________________________
      // do time step


      interpolateToBdry(gfs,initial.iniNS,initial.iniCH, nsV, chV);

      int result = timestepManager.doTimestep(ns_newton,ch_newton,vtkwriter);
      if(result <0)
      {
        return;
      }
      else if(result == 0)
      {
        nsV = nsVOld;
        chV = chVOld;
        continue;
      }

      chVOld = chV;
      nsVOld = nsV;

      //_________________________________________________________________________________________________________________________

      // mark elements for refinement
      typedef RefinementIndicator<GV,RF,ACFS_FEM_AC_Estimator<CH_GFS>,CH_GFS> RIND;
      RIND rInd(gv,gfs.ch,ptree.sub("Refinement"));
      rInd.calculateIndicator(chV);
      rInd.markGrid(grid,2);

      printVector(chV,3,"chV");
      printVector(nsV,3,"nsV");

      // do refinement
      std::cout << "adapt grid and solution" << std::endl;
      auto chTransfer = transferSolutions(gfs.ch,2,chV,chVOld);
      auto nsTransfer = transferSolutions(gfs.ns,2,nsV,nsVOld);
      Dune::PDELab::adaptGrid(grid, chTransfer,nsTransfer);
      std::cout << "adapt grid and solution ended" << std::endl;

      printVector(chV,3,"chV");
      printVector(nsV,3,"nsV");

      // recompute constraints
      gfs.updateConstraintsContainer(bdry);
      interpolateToBdry(gfs,initial.iniNS,initial.iniCH,nsVOld,chVOld);


      vtkwriter.write(timestepManager.t+1e-6,Dune::VTK::ascii);
    }
}
